#include "cglib/arcball_camera.h"
#include <windows.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>

ArcballCamera::ArcballCamera()
{
	mStartRot = Matrix4x4::identity();
	mCurrentRot = Matrix4x4::identity();
}

void ArcballCamera::SetDimensions(int w, int h)
{
    mDimensions[0] = w;
    mDimensions[1] = h;
}

void ArcballCamera::SetPerspective(float fovy)
{
    mPerspective[0] = fovy;
}

void ArcballCamera::SetViewport(int x, int y, int w, int h)
{
    mViewport[0] = x;
    mViewport[1] = y;
    mViewport[2] = w;
    mViewport[3] = h;
    mPerspective[1] = float(w)/h;
	SetDimensions(w, h);
}

void ArcballCamera::SetCenter(const Point3 & center)
{
    mStartCenter = mCurrentCenter = center;
}

void ArcballCamera::SetRotation(const Matrix4x4 & rotation)
{
    mStartRot = mCurrentRot = rotation;
}

void ArcballCamera::SetDistance(const float distance)
{
    mStartDistance = mCurrentDistance = distance;
}

void ArcballCamera::MouseClick(Button button, int x, int y)
{
    mStartClick[0] = x;
    mStartClick[1] = y;	

    mButtonState = button;
    switch (button)
    {
    case LEFT:
        mCurrentRot = mStartRot;
        break;
    case MIDDLE:
        mCurrentCenter = mStartCenter;
        break;
    case RIGHT:
        mCurrentDistance = mStartDistance;
        break;        
    default:
        break;
    }
}

void ArcballCamera::MouseDrag(int x, int y)
{
    switch (mButtonState)
    {
    case LEFT:
        ArcBallRotation(x,y);
        break;
    case MIDDLE:
        PlaneTranslation(x,y);
        break;
    case RIGHT:
        DistanceZoom(x,y);
        break;
    default:
        break;
    }
}

void ArcballCamera::MouseRelease(int x, int y)
{
    mStartRot = mCurrentRot;
    mStartCenter = mCurrentCenter;
    mStartDistance = mCurrentDistance;
    
    mButtonState = NONE;
}


void ArcballCamera::ArcBallRotation(int x, int y)
{
    float sx, sy, sz, ex, ey, ez;
    float scale;
    float sl, el;
    float dotprod;
    
    // find vectors from center of window
    sx = (float)mStartClick[0] - (mDimensions[0]  / 2);
    sy = (float)mStartClick[1] - (mDimensions[1] / 2);
    ex = (float)x -   (mDimensions[0] / 2);
    ey = (float)y -   (mDimensions[1] / 2);
    
    // invert y coordinates (raster versus device coordinates)
    sy = -sy;
    ey = -ey;
    
    // scale by inverse of size of window and magical sqrt2 factor
    if (mDimensions[0] > mDimensions[1]) {
        scale = (float) mDimensions[1];
    } else {
        scale = (float) mDimensions[0];
    }

    scale = 1.0f/scale;
    
    sx *= scale;
    sy *= scale;
    ex *= scale;
    ey *= scale;

    // project points to unit circle
    sl = (float)_hypot(sx, sy);
    el = (float)_hypot(ex, ey);
    
    if (sl > 1.0) {
        sx /= sl;
        sy /= sl;
        sl = 1.0;
    }
    if (el > 1.0) {
        ex /= el;
        ey /= el;
        el = 1.0;
    }
    
    // project up to unit sphere - find Z coordinate
    sz = sqrtf(1.0f - sl * sl);
    ez = sqrtf(1.0f - el * el);
    
    // rotate (sx,sy,sz) into (ex,ey,ez)    
    // compute angle from dot-product of unit vectors (and double
    // it).  compute axis from cross product.
	Vector3 s(sx, sy, sz);
	Vector3 e(ex, ey, ez);
    dotprod = dot(s,e);

    if (dotprod != 1)
	{		
		Vector3 axis = normalize(cross(s,e));
		Matrix4x4 newRotation = Matrix4x4::rotate(2.0f * acos(dotprod) * 180.0f / M_PI, axis);
		mCurrentRot = newRotation * mStartRot;		
	}
    else
    {
        mCurrentRot = mStartRot;
    }
}

void ArcballCamera::PlaneTranslation(int x, int y)
{
    // map window x,y into viewport x,y

    // start
    int sx = mStartClick[0] - mViewport[0];
    int sy = mStartClick[1] - mViewport[1];

    // current
    int cx = x - mViewport[0];
    int cy = y - mViewport[1];


    // compute "distance" of image plane (wrt projection matrix)
    float d = float(mViewport[3])/2.0f / tanf(mPerspective[0]*M_PI / 180.0f / 2.0f);

    // compute up plane intersect of clickpoint (wrt fovy)
    float su = -sy + mViewport[3]/2.0f;
    float cu = -cy + mViewport[3]/2.0f;

    // compute right plane intersect of clickpoint (ASSUMED FOVY is 1)
    float sr = (sx - mViewport[2]/2.0f);
    float cr = (cx - mViewport[2]/2.0f);

    float dr, du;
    dr = cr-sr;
    du = cu-su;
    // this maps move
    dr *= -mCurrentDistance/d;
    du *= -mCurrentDistance/d;

    mCurrentCenter = mStartCenter +
		dr * Vector3(mCurrentRot[0][0],mCurrentRot[1][0],mCurrentRot[2][0]) +
		du * Vector3(mCurrentRot[0][1],mCurrentRot[1][1],mCurrentRot[2][1]);

}

void ArcballCamera::ApplyViewport() const
{
    glViewport(mViewport[0],mViewport[1],mViewport[2],mViewport[3]);
}

void ArcballCamera::ApplyPerspective() const
{
    gluPerspective(mPerspective[0], mPerspective[1], 1.0, 1000.0);   
}


void ArcballCamera::ApplyModelview()
{
    // back up distance
    gluLookAt(0,0,mCurrentDistance,0,0,0,0,1,0);

    // rotate object
	float m[16];
	for(int i=0;i<4;i++)
		for(int j=0;j<4;j++)
		{
			m[i*4 + j] = mCurrentRot[j][i];
		}

	//for(int i=0;i<4;i++)
	//{
		//printf("%.2f %.2f %.2f %.2f\n", 
		//	mCurrentRot[i][0], mCurrentRot[i][1], mCurrentRot[i][2], mCurrentRot[i][3]);		
	//}
	//printf("\n");

    glMultMatrixf(m);
    
    //translate object to center
    glTranslatef(-mCurrentCenter[0],-mCurrentCenter[1],-mCurrentCenter[2]);    
}

void ArcballCamera::DistanceZoom(int x, int y)
{
    int sy = mStartClick[1] - mViewport[1];
    int cy = y - mViewport[1];

    double delta = double(cy-sy)/mViewport[3];

    // exponential zoom factor
    mCurrentDistance = mStartDistance * (float)exp(delta);  
}