#define NOMINMAX
#include "cglib/parser.h"
#include "cglib/object.h"
#include "cglib/reference.h"
#include "cglib/string.h"
#include "cglib/path.h"
#include <map>
#include <fstream>
#include <sstream>
#include <iostream>
#include <limits>
#include "cglib/mesh.h"
#include "cglib/mesh_material_mapping.h"
#include "cglib/material_list.h"
#include "cglib/phong_material.h"
#include "cglib/textured_phong_material.h"
#include "cglib/texture.h"
#include "cglib/mesh_node.h"
#include "cglib/transform_node.h"
#include "cglib/group_node.h"
#include "cglib/rgb.h"

#define FOR(__i, __times) for(int __i=0;__i<__times;__i++)

class Parser::ParserImplementation
{
public:
    ParserImplementation();
    virtual ~ParserImplementation();
	
	Object *parse(std::string &file_name);
	
private:
    std::string directory_name_; //!< the path to the directory containing the output file
	std::string file_name_; //! the input file name
    std::map< std::string, Reference<Object> > object_map_; //! a map fron UUID to the object associated with it
	std::ifstream input_file_; //!< the input file.
	
	//! Parse the input file and return an Object.
	Object *parse_Object();
	
	typedef Object *(ParserImplementation::*FactoryMethod)();
	std::map< std::string, FactoryMethod > factories_;
	void init_factories();
	
	typedef void (ParserImplementation::*ParseContentMethod)(Object *object);
	std::map< std::string, ParseContentMethod > content_parsers_;
	void init_content_parsers();
	
	// FACTORY METHODS
	Object *create_Mesh();
	Object *create_MeshMaterialMapping();
	Object *create_MaterialList();
	Object *create_PhongMaterial();
	Object *create_TexturedPhongMaterial();
	Object *create_MeshNode();
	Object *create_TransformNode();
	Object *create_GroupNode();
	Object *create_Texture();
	
	// CONTENT PARSER METHODS
	void parse_Mesh_content(Object *object);
	void parse_MeshMaterialMapping_content(Object *object);
	void parse_MaterialList_content(Object *object);
	void parse_PhongMaterial_content(Object *object);
	void parse_TexturedPhongMaterial_content(Object *object);
	void parse_MeshNode_content(Object *object);
	void parse_GroupNode_content(Object *object);
	void parse_TransformNode_content(Object *object);
	void parse_Texture_content(Object *object);
	
	// HELPER METHODS
    void go_to_new_line();
    std::string get_word();
    std::string get_line();
    int get_int();
    int get_int_line();
    float get_float();
    float get_float_line();
    Point3 get_Point3();
    Vector3 get_Vector3();
    Uv get_Uv();
    Matrix4x4 get_Matrix4x4();
    Transform get_Transform();
    Rgba get_Rgba();
    Rgb get_Rgb();
    
    Object *create_new_object(std::string &class_name);
	void check_class_name_exists(std::string &class_name);
	void register_object_by_uuid(Object *object, std::string &uuid);
	void parse_content(Object *object, std::string &class_name);
};

void Parser::ParserImplementation::init_factories()
{
    factories_["Mesh"] = &Parser::ParserImplementation::create_Mesh;
    factories_["MeshMaterialMapping"] = &Parser::ParserImplementation::create_MeshMaterialMapping;
    factories_["MaterialList"] = &Parser::ParserImplementation::create_MaterialList;
    factories_["PhongMaterial"] = &Parser::ParserImplementation::create_PhongMaterial;
    factories_["TexturedPhongMaterial"] = &Parser::ParserImplementation::create_TexturedPhongMaterial;
    factories_["MeshNode"] = &Parser::ParserImplementation::create_MeshNode;
    factories_["TransformNode"] = &Parser::ParserImplementation::create_TransformNode;
    factories_["GroupNode"] = &Parser::ParserImplementation::create_GroupNode;
    factories_["Texture"] = &Parser::ParserImplementation::create_Texture;
}

void Parser::ParserImplementation::init_content_parsers()
{
    content_parsers_["Mesh"] = &Parser::ParserImplementation::parse_Mesh_content;
    content_parsers_["MeshMaterialMapping"] = &Parser::ParserImplementation::parse_MeshMaterialMapping_content;
    content_parsers_["MaterialList"] = &Parser::ParserImplementation::parse_MaterialList_content;
    content_parsers_["PhongMaterial"] = &Parser::ParserImplementation::parse_PhongMaterial_content;
    content_parsers_["TexturedPhongMaterial"] = &Parser::ParserImplementation::parse_TexturedPhongMaterial_content;
    content_parsers_["TransformNode"] = &Parser::ParserImplementation::parse_TransformNode_content;
    content_parsers_["GroupNode"] = &Parser::ParserImplementation::parse_GroupNode_content;
	content_parsers_["MeshNode"] = &Parser::ParserImplementation::parse_MeshNode_content;
    content_parsers_["Texture"] = &Parser::ParserImplementation::parse_Texture_content;
}

Parser::Parser()
{
    self = new ParserImplementation();
}

Parser::~Parser()
{
    delete self;        
}

Parser::ParserImplementation::ParserImplementation()
{
    init_factories();
    init_content_parsers();
}

Parser::ParserImplementation::~ParserImplementation()
{
}

Object *Parser::parse(std::string &file_name)
{
	return self->parse(file_name);
}

Object *Parser::ParserImplementation::parse(std::string &file_name)
{
	object_map_.clear();
	file_name_ = file_name;
	directory_name_ = Path::v()->get_directory_name(file_name);
	input_file_.open(file_name.c_str());

	Object *result = parse_Object();

	input_file_.close();

	return result;
}

Object *Parser::ParserImplementation::parse_Object()
{
	std::string class_name = get_word();
	std::string uuid = get_word();
	go_to_new_line();
	
	if (class_name == "Reference") 
	{
	    Object *result = object_map_[uuid].get_pointer();
	    return result;
	}
	else 
	{
        Object *result = create_new_object(class_name);
    	register_object_by_uuid(result, uuid);
    	parse_content(result, class_name);
    	return result;
	}
}

Object *Parser::ParserImplementation::create_new_object(std::string &class_name)
{
    check_class_name_exists(class_name);
	FactoryMethod factory = factories_[class_name];
	return (this->*factory)();        
}

void Parser::ParserImplementation::check_class_name_exists(std::string &class_name)
{
	if (factories_.find(class_name) == factories_.end())
	{
		stringstream error;
		error << "EasyTextParser::parse()" 
			<< " : class " << class_name << " does not exist";
		throw std::exception(error.str().c_str());
	}
}

void Parser::ParserImplementation::register_object_by_uuid(Object *object, std::string &uuid) 
{
    Reference< Object > ref = object;
    object_map_[uuid] = ref;    
}

void Parser::ParserImplementation::parse_content(Object *object, std::string &class_name) 
{
    ParseContentMethod parse_content = content_parsers_[class_name];
	(this->*parse_content)(object);    
}

void Parser::ParserImplementation::go_to_new_line()
{
	input_file_.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
}

std::string Parser::ParserImplementation::get_word()
{
	std::string result;
	input_file_ >> result;
	return result;
}			

int Parser::ParserImplementation::get_int_line()
{
	int result = get_int();
	go_to_new_line();
	return result;
}

int Parser::ParserImplementation::get_int()
{
	int result;
	input_file_ >> result;
	return result;
}

Point3 Parser::ParserImplementation::get_Point3()
{
	float x = get_float();
	float y = get_float();
	float z = get_float();
	go_to_new_line();
	return Point3(x,y,z);
}

float Parser::ParserImplementation::get_float()
{
	float result;
	input_file_ >> result;
	return result;
}

Vector3 Parser::ParserImplementation::get_Vector3()
{
	float x = get_float();
	float y = get_float();
	float z = get_float();
	go_to_new_line();
	return Vector3(x,y,z);
}

Uv Parser::ParserImplementation::get_Uv()
{
	float u = get_float();
	float v = get_float();
	go_to_new_line();
	return Uv(u,v);
}

Rgba Parser::ParserImplementation::get_Rgba()
{
	float r = get_float();
	float g = get_float();
	float b = get_float();
	float a = get_float();
	go_to_new_line();
	return Rgba(r,g,b,a);
}

float Parser::ParserImplementation::get_float_line()
{
	float result = get_float();
	go_to_new_line();
	return result;
}

Transform Parser::ParserImplementation::get_Transform()
{
	Matrix4x4 m = get_Matrix4x4();
	Matrix4x4 mi = get_Matrix4x4();
	Matrix4x4 mit = get_Matrix4x4();
	return Transform(m, mi, mit);
}

Matrix4x4 Parser::ParserImplementation::get_Matrix4x4()
{
	float m00 = get_float();
	float m01 = get_float();
	float m02 = get_float();
	float m03 = get_float();
	
	float m10 = get_float();
	float m11 = get_float();
	float m12 = get_float();
	float m13 = get_float();

	float m20 = get_float();
	float m21 = get_float();
	float m22 = get_float();
	float m23 = get_float();

	float m30 = get_float();
	float m31 = get_float();
	float m32 = get_float();
	float m33 = get_float();

	go_to_new_line();

	return Matrix4x4(
		m00, m01, m02, m03,
		m10, m11, m12, m13,
		m20, m21, m22, m23,
		m30, m31, m32, m33);
}

Rgb Parser::ParserImplementation::get_Rgb()
{
	float r = get_float();
	float g = get_float();
	float b = get_float();
	go_to_new_line();
	return Rgb(r,g,b);
}

std::string Parser::ParserImplementation::get_line()
{
	std::string line;
	std::getline(input_file_, line);
	return line;
}


// FACTORY METHODS
Object *Parser::ParserImplementation::create_Mesh()
{
    return new Mesh();
}

Object *Parser::ParserImplementation::create_MeshMaterialMapping()
{
    return new MeshMaterialMapping();
}

Object *Parser::ParserImplementation::create_MaterialList()
{
    return new MaterialList();
}

Object *Parser::ParserImplementation::create_PhongMaterial()
{
    return new PhongMaterial();
}

Object *Parser::ParserImplementation::create_TexturedPhongMaterial()
{
    return new TexturedPhongMaterial();
}

Object *Parser::ParserImplementation::create_MeshNode()
{
    return new MeshNode();
}

Object *Parser::ParserImplementation::create_TransformNode()
{
    return new TransformNode();
}

Object *Parser::ParserImplementation::create_GroupNode()
{
    return new GroupNode();
}

Object *Parser::ParserImplementation::create_Texture()
{
    return new Texture();
}

// CONTENT PARSER METHODS
void Parser::ParserImplementation::parse_Mesh_content(Object *object)
{
    Mesh *mesh = dynamic_cast<Mesh *>(object);
    
    int position_count = get_int_line();
    FOR(position_index, position_count)
        mesh->append_position(get_Point3());
        
    int normal_count = get_int_line();
    FOR(normal_index, normal_count)
        mesh->append_normal(get_Vector3());
        
    int tex_coord_count = get_int_line();
    FOR(tex_coord_index, tex_coord_count)
        mesh->append_tex_coord(get_Uv());
        
    int face_count = get_int_line();
    FOR(face_index, face_count) 
    {
        int vertex_count = get_int();
        mesh->append_new_face();
        FOR(vertex_index, vertex_count) 
        {
            int position_index = get_int();
            int normal_index = get_int();
            int tex_coord_index = get_int();
            mesh->append_vertex_to_last_face(position_index, normal_index, tex_coord_index);
        }
		go_to_new_line();
    }
}

void Parser::ParserImplementation::parse_MeshMaterialMapping_content(Object *object)
{
    MeshMaterialMapping *mapping = dynamic_cast<MeshMaterialMapping *>(object);
    
    int mapping_count = get_int_line();
    FOR(mapping_index, mapping_count)
        mapping->append_face_material_index(get_int_line());
}

void Parser::ParserImplementation::parse_MaterialList_content(Object *object)
{
    MaterialList *material_list = dynamic_cast<MaterialList *>(object);
    
    int material_count = get_int_line();
    FOR(material_index, material_count)
    {
        Material *material = dynamic_cast<Material *>(parse_Object());
        material_list->append_material(material);           
    }
}

void Parser::ParserImplementation::parse_PhongMaterial_content(Object *object)
{
    PhongMaterial *phong = dynamic_cast<PhongMaterial *>(object);
    
    phong->ambient = get_Rgba();
    phong->diffuse = get_Rgba();
    phong->specular = get_Rgba();
    phong->shininess = get_float_line();
}

void Parser::ParserImplementation::parse_TexturedPhongMaterial_content(Object *object)
{
    TexturedPhongMaterial *phong = dynamic_cast<TexturedPhongMaterial *>(object);
    
    phong->ambient = get_Rgba();
    phong->diffuse = get_Rgba();
    phong->specular = get_Rgba();
    phong->shininess = get_float_line();
    phong->texture = dynamic_cast<Texture *>(parse_Object());    
}

void Parser::ParserImplementation::parse_MeshNode_content(Object *object)
{
    MeshNode *mesh_node = dynamic_cast<MeshNode *>(object);
    
    Mesh *mesh = dynamic_cast<Mesh *>(parse_Object());
    MeshMaterialMapping *material_mapping = dynamic_cast<MeshMaterialMapping *>(parse_Object());
    MaterialList *material_list = dynamic_cast<MaterialList *>(parse_Object());
    
    mesh_node->set_mesh(mesh);
    mesh_node->set_material_mapping(material_mapping);
    mesh_node->set_material_list(material_list);
}

void Parser::ParserImplementation::parse_TransformNode_content(Object *object)
{
    TransformNode *transform_node = dynamic_cast<TransformNode *>(object);
    
    transform_node->set_transform(get_Transform());
    SceneGraphNode *child = dynamic_cast<SceneGraphNode *>(parse_Object());
    transform_node->set_child(child);
}

void Parser::ParserImplementation::parse_GroupNode_content(Object *object)
{
    GroupNode *group_node = dynamic_cast<GroupNode *>(object);
    
    int child_count = get_int_line();
    FOR(child_index, child_count) 
    {
        SceneGraphNode *child = dynamic_cast<SceneGraphNode *>(parse_Object());
        group_node->append_child(child);               
    }
}

void Parser::ParserImplementation::parse_Texture_content(Object *object)
{
    Texture *texture = dynamic_cast<Texture *>(object);
    
    std::string file_name = strip(get_line());
    file_name = Path::v()->join(directory_name_, file_name);
    texture->load(file_name.c_str());
}