#ifndef _POINT3_H_
#define _POINT3_H_

#include "cglib/vector3.h"

//! Represents a point in 3D.
struct Point3 
{
public:
    //! The x coordinate.
	float x;
	//! The y coordinate.
	float y;
	//! The z coordinate.
	float z;
	
	//! Initializes all components to zero.
	Point3(): x(0), y(0), z(0) 
	{
	}
	
	//! Initializes all components to the given real number \a c.
	Point3(float c): x(c), y(c), z(c) 
	{
	}
	
	//! Initializes all components to the corresponding parameters.
	Point3(float _x, float _y, float _z) : x(_x), y(_y), z(_z) 
	{
	}

    //! Add the given vector to \a this and return the result (which is a point).
	Point3 operator+(const Vector3 &v) const 
	{
		return Point3(x + v.x, y + v.y, z + v.z);
	}
	
	//! Add the given vector to \a this and set \a this to the result.
	Point3 &operator+=(const Vector3 &v) 
	{
		x += v.x; y += v.y; z += v.z;
		return *this;
	}

    //! Subtract the given point from \a this and return the result.
	Vector3 operator-(const Point3 &p) const 
	{
		return Vector3(x - p.x, y - p.y, z - p.z);
	}
	
	//! Subtract the given vector from \a this and return the result.
	Point3 operator-(const Vector3 &v) const 
	{
		return Point3(x - v.x, y - v.y, z - v.z);
	}
	
	//! Subtract the given vector from \a this and set \a this to the result.
	Point3 &operator-=(const Vector3 &v) 
	{
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}
	
	//! Check if the given point to \a this.
	bool operator==(const Point3 &other) const
	{
		return x == other.x && y == other.y && z == other.z;
	}

    //! Check if the given point is not equal to \a this.
	bool operator!=(const Point3 &other) const
	{
		return x != other.x || y != other.y || z != other.z;
	}

    //! Return the <i>i</i>th component of the point.
    /*!
        The components are indexed as follows:
        * The 0th component is the x coordinate.
        * The 1st component is the y coordinate.
        * The 2nd component is the z coordinate.
     */
	float operator[](int i) const 
	{ 
		return (&x)[i]; 
	}
	
	//! Return the reference of the <i>i</i>th component of the point.
	/*!
	    This is done so that you can do this:
	    \code
	        Point3 p;
	        p[0] = 1;
	        p[1] = 2;
	        p[2] = 3;
	    \endcode
	    
	    The components are indexed as follows:
        * The 0th component is the x coordinate.
        * The 1st component is the y coordinate.
        * The 2nd component is the z coordinate.
	 */
	float &operator[](int i) 
	{ 
		return (&x)[i]; 
	}
	
	//! Set the elements of the given array \a a to the cooridnates of the points.
	/*!
	     That is:
	    - \c a[0] will be set to the x coordinate.
	    - \c a[1] will be set to the y coordinate.
	    - \c a[2] will be set to the z coordinate.
	 */
	void fill_array_3(float *a)
	{
		a[0] = x;
		a[1] = y;
		a[2] = z;
	}

    //! Set the elements of the given array \a a to the homogeneous cooridnates of the points.
	/*!
	     That is:
	    - \c a[0] will be set to the x coordinate.
	    - \c a[1] will be set to the y coordinate.
	    - \c a[2] will be set to the z coordinate.
	    - \c a[3] will be set to 1.
	 */
	void fill_array_4(float *a)
	{
		a[0] = x;
		a[1] = y;
		a[2] = z;
		a[3] = 1.0f;
	}
	
	//! Check if each coordinate of \a this is within \a threshold (a real number) of the corresponding coordinate of \a other.
    /*!
        Use this method to test approximate equality.
     */
	bool is_near(const Point3 &other, float threshold=EPSILON) const
	{
		float absx = fabs(x-other.x);
		float absy = fabs(y-other.y);
		float absz = fabs(z-other.z);
		return (absx < EPSILON && absy < EPSILON && absz < EPSILON);
	}
};

//! Return the distance between the two given points.
inline float distance(const Point3 &p1, const Point3 &p2) 
{
	return (p1 - p2).length();
}

//! Return the square of the distance between the two given points.
inline float distance_squared(const Point3 &p1, const Point3 &p2) 
{
	return (p1 - p2).length_squared();
}

#endif /* _POINT3_H_ */
