#ifndef _REFERENCE_H_
#define _REFERENCE_H_

#include <stdlib.h>

//! A container for reference-counted objects.
/*!
    Reference holds a pointer to objects of class
    ReferenceCounted. When you assign such a pointer
    to Reference, the reference count is automatically
    increased. Also, when the pointer Reference holds
    changes, Reference automatically decreases the
    reference count of the old pointer, thereby
    facilitating automatic destruction of the object.
    
    Reference behaves like other "smart pointer" objects
    in C++ such as \c boost::shared_ptr. You assign,
    compare, and use the \c -> operator on Reference.
 */
template <class T>
class Reference
{
private:
	T *pointer_; //!< the pointer to a reference-counted object.

public:
    //! Create a reference to the object pointed to by the given pointer.
    /*!
        \param _pointer a pointer to a reference-counted object or NULL. It is defaulted to NULL.
     */
	inline Reference(T *_pointer = NULL)
	{
		if (_pointer != NULL)
		{
			_pointer->inc_ref();
		}
		pointer_ = _pointer;
	}

    //! A copy constructor.
    /*!
        Notes that this constructor increases the reference count of the object
        \a other holds by one.
     */
	inline Reference(const Reference<T> &other)
	{
		pointer_ = other.pointer_;
		if (pointer_ != NULL)
			pointer_->inc_ref();
	}

    //! Assigns the pointer this instance of Reference holds.
    /*!
        The method also automatically decreases the reference
        of the old object held if it is neither NULL nor
        equal to to the new object being assigned.
     */
	inline Reference &operator=(T *_ptr)
	{				
		if (_ptr != pointer_)
		{
			if (pointer_ != NULL)
			{
				pointer_->dec_ref();				
			}

			pointer_ = _ptr;
			if (pointer_ != NULL)
			{
				pointer_->inc_ref();
			}			
		}
		return *this;
	}

    //! Have \a this reference the object held by \a other.
	inline Reference &operator=(const Reference<T> &other)
	{
		*this = other.pointer_;
		return *this;
	}

    //! Return the underlying pointer.
    /*!
        Use this method to access the underlying object's fields
        or invoking methods.
     */
	inline T *operator->() const
	{
		return pointer_;
	}

    //! Convert the the reference to a boolean, which is true if the underlying object is not NULL.
	inline operator bool() const
	{
		return pointer_ != NULL;
	}

    //! Check whether \a this holds the given pointer.
	inline bool operator==(T *_ptr) const
	{
		return pointer_ == _ptr;
	}

    //! Check whether \a this holds a difference pointer than the given pointer.
	inline bool operator!=(T *_ptr) const
	{
		return pointer_ != _ptr;
	}

    //! Check whether \a this holds the same object as \a other does.
	inline bool operator==(const Reference<T> &other) const
	{
		return pointer_ == other.pointer_;
	}

    //! Check whether \a this holds a difference object than \a other does.
	inline bool operator!=(const Reference<T> &other_) const
	{
		return pointer_ != other.pointer_;
	}

    //! Compere \a this with the given pointer.
	inline bool operator<(T *_ptr) const
	{
		return pointer_ < _ptr;
	}

    //! Compare the pointer \a this holds with the one \a other holds.
	inline bool operator<(const Reference<T> &other) 
	{
		return pointer_ < other.pointer;
	}

    //! Destroy the reference by setting the pointer it holds to NULL.
    /*!
        This would cause the reference count to the held
        object to be decremented, thereby facilitating 
        automatic desctruction of the held object.
     */
	virtual ~Reference()
	{
		if (pointer_ != NULL)
		{
			pointer_->dec_ref();
			pointer_ = NULL;
		}
	}

    //! Return the underlying pointer.
	inline T *get_pointer()
	{
		return pointer_;
	}
};

//! Check whether the given pointer is equal to the pointer the given reference holds.
template <class T>
inline bool operator==(T *_ptr, const Reference<T> ref)
{
	return ref == _ptr;
}

//! Check whether the given pointer is not equal to the pointer the given reference holds.
template <class T>
inline bool operator!=(T *_ptr, const Reference<T> ref)
{
	return ref != _ptr;
}

//! Check whether the given pointer is less than the pointer the given reference holds.
template <class T>
inline bool operator<(T *_ptr, const Reference<T> ref)
{
	return ref < _ptr;
}

#endif /* _REFERENCE_H_ */
