#ifndef _REFERENCE_COUNTED_H_
#define _REFERENCE_COUNTED_H_

//! Base classe for reference-counted objects.
/*!
    It is strongly advised that reference-counted objects only
    be created on the heap. If they are created on the stack,
    you should not call \c dec_ref; otherwise the program will
    crash.
 */
class ReferenceCounted
{
private:
	int ref_count_; //!< The number of references to the object.

public:
    //! Initialize the reference count to zero.
	ReferenceCounted() 
	{
		ref_count_ = 0;
	}

    //! A constructor that does nothing.
	virtual ~ReferenceCounted()
	{
	}

    //! Increase the reference by one.
	inline void inc_ref()
	{
		ref_count_++;
	}

    //! Decrease the reference by one. Deallocate the object from the heap if the reference count is zero.
    /*!
        This function should not be called if the object
        is created on the stack.
     */
	inline void dec_ref()
	{
		ref_count_--;
		if (ref_count_ <= 0)
			delete this;
	}

    //! Return the number of references to the object.
	inline int get_ref_count()
	{
		return ref_count_;
	}
};

#endif /* _REFERENCE_COUNTED_H_ */
