#ifndef _RGBA_H_
#define _RGBA_H_

#include "cglib/definitions.h"

//! Represents an RGBA color.
struct Rgba
{
public:
	float r; //!< the intensity of the color red
	float g; //!< the intensity of the color green
	float b; //!< the intensity of the color blue
	float a; //!< the transparency.

    //! Initialize all components to zero.
	inline Rgba()
	{
		r = g = b = a = 0.0f;
	}

    //! Initialize all components to the given real number \a c.
	inline Rgba(float c)
	{
		r = g = b = a = c;
	}

    //! Initialize the colors to the corresponding given values and the alpha to 1.
	inline Rgba(float _r, float _g, float _b)
	{
		r = _r;
		g = _g;
		b = _b;
		a = 1.0f;
	}

    //! Initialize all components to the corresponding given values.
	inline Rgba(float _r, float _g, float _b, float _a)
	{
		r = _r;
		g = _g;
		b = _b;
		a = _a;
	}

    //! Add the given RGBA \a other to \a this and return the result.
	inline Rgba operator+(const Rgba &other) const {
		return Rgba(r + other.r, g + other.g, b + other.b, a + other.a);
	}

    //! Subtract the given RGBA \a other from \a this and return the result.
	inline Rgba operator-(const Rgba &other) const {
		return Rgba(r - other.r, g - other.g, b - other.b, a - other.a);
	}

    //! Add the given real number \a c to every component of \a this and return the result.
	inline Rgba operator+(float c) const {
		return Rgba(r + c, g + c, b + c, a + c);
	}

    //! Add the given RGBA \a other to \a this and set \a this to the result.
	inline Rgba &operator+=(const Rgba &other)
	{
		r += other.r;
		g += other.g;
		b += other.b;
		a += other.a;
		return *this;
	}

    //! Component-wisely multiply the given RGBA \a other to \a this and return the result.
	inline Rgba operator*(const Rgba &other) const {
		return Rgba(r * other.r, g * other.g, b * other.b, a * other.a);
	}

    //! Multiply the given real number \a c with \a this and return the result.
	inline Rgba operator*(float c) const {
		return Rgba(r * c, g * c, b * c, a * c);
	}

    //! Component-wisely multiply the given RGB \a other to \a this and then set \a this to the result.
	inline Rgba &operator*=(const Rgba &other)
	{
		r *= other.r;
		g *= other.g;
		b *= other.b;
		a *= other.a;
		return *this;
	}

    //! Divide all the components of \a this with the corresponding components of \a other and return the result.
	inline Rgba operator/(const Rgba &other) const {
		return Rgba(r / other.r, g / other.g, b / other.b, a / other.a);
	}

    //! Divide all the components of \a this with the given real number \a c and return the result.
	inline Rgba operator/(float c) const {
		float cinv = 1.0f / c;
		return Rgba(r * cinv, g * cinv, b * cinv, a * cinv);
	}

    //! Fill the given array with the values of the components of \a this.
    /*!
        That is:
        * \c a[0] is set to \a r.
        * \c a[1] is set to \a g.
        * \c a[2] is set to \a b.
        * \c a[3] is set to \a a.
     */
	inline void fill_array_4(float *_a)
	{
		_a[0] = r;
		_a[1] = g;
		_a[2] = b;
		_a[3] = a;
	}

    //! Check whether all the components of \a this is within \a threshold of the corresponding components of \a other or not.
    /*!
        Use this method to test approximate equality.
     */
	bool is_near(const Rgba &other, float threshold=EPSILON)
	{
		float absr = fabs(r - other.r);
		float absg = fabs(g - other.g);
		float absb = fabs(b - other.b);
		float absa = fabs(a - other.a);
		return (absr < EPSILON && absg < EPSILON && absb < EPSILON && absa < EPSILON);
	}
};

//! Add the given real number \a c to all the components of the given RGBA and then return the result.
inline Rgba operator+(float c, const Rgba &rgba)
{
	return Rgba(rgba.r + c, rgba.g + c, rgba.b + c, rgba.a + c);
}

//! Multiply the given real number \a to all the components of the given RGBA and then return the result.
inline Rgba operator*(float c, const Rgba &rgba)
{
	return Rgba(rgba.r * c, rgba.g * c, rgba.b * c, rgba.a * c);
}

#endif /* _RGBA_H_ */
