#ifndef _TRANSFORM_H_
#define _TRANSFORM_H_

#include "cglib/matrix4x4.h"

//! Represent an affine transformation.
struct Transform
{
public:
	Matrix4x4 m;   //!< the matrix of the transformation
	Matrix4x4 mi;  //!< the inverse of the matrix of the transformation
	Matrix4x4 mit; //!< the inverse transpose of the matrix of the transformation

    //! Create an identity transformation.
	Transform();
	
	//! Create a transformation with the given matrix and its inverse.
	Transform(const Matrix4x4 &_m, const Matrix4x4 &_mi);
	
	//! Create a transformation with the given matrix, its inverse, and its inverse transpose.
	Transform(const Matrix4x4 &_m, const Matrix4x4 &_mi, const Matrix4x4 &_mit);

    //! Multiply \a this with the given transformation, thereby composing the two, and return the result.
	Transform operator*(const Transform &other) const;
	
	//! Multiply \a this with the given transformation, thereby composing the two, and set \a this to the result.
	Transform &operator*=(const Transform &otherz);

	//! Check whether the given transform \a other is within \a threshold of \a this.
	bool is_near(const Transform &other, float threshold=0.0001f);

    //! Create an identity transformation.
	static Transform identity();

    //! Create a translation.
	static Transform translate(float x, float y, float z);
	
	//! Create a translation along the x-axis.
	static Transform translate_x(float x);
	
	//! Create a translation along the y-axis.
	static Transform translate_y(float y);
	
	//! Create a translation along the z-axis.
	static Transform translate_z(float z);

    //! Create a scaling.
	static Transform scale(float x, float y, float z);
	
	//! Create a scaling with the same scaling factor \a s along all axes.
	static Transform scale(float s);
	
	//! Create a scaling along the x-axis.
	static Transform scale_x(float x);
	
	//! Create a scaling along the y-axis.
	static Transform scale_y(float y);
	
	//! Create a scaling along the z-axis.
	static Transform scale_z(float z);

    //! Create a rotation, specifying the degrees to rotate and the axis of rotation.
	static Transform rotate(float degrees, Vector3 axis);
	
	//! Create a rotation around the x-axis.
	static Transform rotate_x(float degrees);
	
	//! Create a rotation around the y-axis.
	static Transform rotate_y(float degrees);
	
	//! Create a rotation around the z-axis.
	static Transform rotate_z(float degrees);
};


#endif /* _TRANSFORM_H_ */
