/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

#ifndef _VECTOR3_H_
#define _VECTOR3_H_

#include "cglib/definitions.h"

struct Point3;

//! Represents a vector in 3D.
struct Vector3 
{
public:
    //! The x coordinate.
    float x; 
    //! The y coordinate.
	float y; 
	//! The z coordinate.
	float z; 

	//! Initializes all coordinates to 0.
	Vector3(): x(0), y(0), z(0) 
	{
	}
	
	//! Initializes all coordinates to the given value \a c. 
	Vector3(float c): x(c), y(c), z(c) 
	{
	}
	
	//! Initializes each coordinate to the corresponding given values.
	/*!
	    \param _x the initial x coordinate
	    \param _y the initial y coordinate
	    \param _z the initial z coordinate
	 */
	Vector3(float _x, float _y, float _z)
		: x(_x), y(_y), z(_z) 
	{
	}
	
	//! Initialize each coordinate to the corresponding cooridnate of the given point \a p.
	explicit Vector3(const Point3 &p);

    //! Add \a this and the given vector \a v together and then return the result.
	Vector3 operator+(const Vector3 &v) const 
	{
		return Vector3(x + v.x, y + v.y, z + v.z);
	}
	
	//! Add \a this and the given vector \a v together and set \a this to the result.
	Vector3& operator+=(const Vector3 &v) 
	{
		x += v.x; y += v.y; z += v.z;
		return *this;
	}

    //! Subtract \a this with \a v and return the result.
	Vector3 operator-(const Vector3 &v) const 
	{
		return Vector3(x - v.x, y - v.y, z - v.z);
	}
	
	//! Subtract \a this with \a v and set \a this to the result.
	Vector3& operator-=(const Vector3 &v) 
	{
		x -= v.x; y -= v.y; z -= v.z;
		return *this;
	}
	
	//! Check if \a this is equal to \a v or not.
	bool operator==(const Vector3 &v) const 
	{
		return x == v.x && y == v.y && z == v.z;
	}
	
	//! Multiply \a this with the given real number \a f and return the result.
	Vector3 operator*(float f) const 
	{
		return Vector3(f*x, f*y, f*z);
	}
	
	//! Multiply \a this with the given real number \a f and set \a this to the result. 
	Vector3 &operator*=(float f) 
	{
		x *= f; y *= f; z *= f;
		return *this;
	}
	
	//! Divide \a this by the given real number \a f (which must not be zero) and return the result.
	Vector3 operator/(float f) const 
	{
		float inv = 1.f / f;
		return Vector3(x * inv, y * inv, z * inv);
	}
	
	//! Divde \a this by the given real number \a f (which must not be zero) and set \a this to the result.
	Vector3 &operator/=(float f) 
	{
		float inv = 1.f / f;
		x *= inv; y *= inv; z *= inv;
		return *this;
	}
	
	//! Return the negative of \a this.
	Vector3 operator-() const 
	{
		return Vector3(-x, -y, -z);
	}
	
	//! Return the <i>i</i>th component of \a this.
	/*!
	    The components are indexed as follows:
	    - The 0th component is x.
        - The 1st component is y.
        - The 2nd component is z.
	 */
	float operator[](int i) const 
	{
		return (&x)[i];
	}
	
	//! Check if \a this is not equal to the given vector \a other.
	bool operator!=(const Vector3 &other) const
	{
		return !(*this == other);
	}
	
	//! Return the reference to the <i>i</i>th component of \a this.
	/*!
	    This method is provided so that you can do something like this:
	    \code 
	        Vector3 a;
	        a[0] = 1.0f;
	        a[1] = 1.5f;
	        a[2] = 2.0f;
	    \endcode
    	
	    The components are indexed as follows:
	    - The 0th component is x.
        - The 1st component is y.
        - The 2nd component is z.	    
	 */
	float &operator[](int i) 
	{
		return (&x)[i];
	}
	
	//! Fill the given 3-element array \a with the vector's coordinates.
	/*!
	    That is:
	    - \c a[0] will be set to the x coordinate.
	    - \c a[1] will be set to the y coordinate.
	    - \c a[2] will be set to the z coordinate.
	 */
	void fill_array_3(float *a)
	{
		a[0] = x;
		a[1] = y;
		a[2] = z;
	}
	
	//! Fill the given 4-element array \a with the vector's homogeneous coordinates.
	/*!
	    That is:
	    - \c a[0] will be set to the x coordinate.
	    - \c a[1] will be set to the y coordinate.
	    - \c a[2] will be set to the z coordinate.
	    - \c a[3] will be set to 0.
	 */
	void fill_array_4(float *a)
	{
		a[0] = x;
		a[1] = y;
		a[2] = z;
		a[3] = 0.0f;
	}
	
	//! Return the square of the length of \a this.
	float length_squared() const 
	{ 
		return x*x + y*y + z*z; 
	}
	
	//! Return the length of \a this.
	float length() const 
	{ 
		return sqrtf(x*x + y*y + z*z); 
	}

    //! Check if each coordinate of \a this is within \a threshold (a real number) of the corresponding coordinate of \a other.
    /*!
        Use this method to test approximate equality.
     */
	bool is_near(const Vector3 &other, float threshold=EPSILON) const
	{
		float absx = fabs(x-other.x);
		float absy = fabs(y-other.y);
		float absz = fabs(z-other.z);
		return (absx < EPSILON && absy < EPSILON && absz < EPSILON);
	}
};	

//! Multifly the vector \a v with the given real number \a f and return the result.
inline Vector3 operator*(float f, const Vector3 &v) 
{
	return v*f;
}

//! Return the dot product of the given two vectors.
inline float dot(const Vector3 &v1, const Vector3 &v2) 
{
	return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

//! Return the absolute value of the dot product of the given two vectors.
inline float absdot(const Vector3 &v1, const Vector3 &v2) 
{
	return fabsf(dot(v1, v2));
}

//! Return the cross product of the given two vectors. 
inline Vector3 cross(const Vector3 &v1, const Vector3 &v2) 
{
	return Vector3((v1.y * v2.z) - (v1.z * v2.y),
				  (v1.z * v2.x) - (v1.x * v2.z),
				  (v1.x * v2.y) - (v1.y * v2.x));
}

//! Return a unit vector pointing in the same direction as the given vector \a v.
inline Vector3 normalize(const Vector3 &v) 
{
	return v / v.length();
}

#endif /* _VECTOR3_H_ */
